/*
*	File:		StereoImageMunger.cpp
*	
*	Version:	1.0.2
* 
*	Created:	2005/3/24
*	
*	Copyright:  Copyright © 2005-2008 Paul Cantrell
*
*  http://innig.net/
*  
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
* 
*  (1) Redistributions of source code must retain the above copyright
*      notice, this list of conditions and the following disclaimer.
* 
*  (2) Redistributions in binary form must reproduce the above copyright
*      notice, this list of conditions and the following disclaimer in
*      the documentation and/or other materials provided with the
*      distribution. 
* 
*  (3) The name of the author may not be used to endorse or promote
*      products derived from this software without specific prior
*      written permission.
* 
*  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
*  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
*  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
*  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
*  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
*  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
*  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
*  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
*  OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*=============================================================================
	StereoImageMunger.cpp
	
=============================================================================*/
#include "StereoImageMunger.h"


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

COMPONENT_ENTRY(StereoImageMunger)


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//	StereoImageMunger::StereoImageMunger
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
StereoImageMunger::StereoImageMunger(AudioUnit component)
	: AUEffectBase(component)
    {
	CreateElements();
	Globals()->UseIndexedParameters(kNumberOfParameters);
	SetParameter(kParam_In1, kDefaultValue_In1 );
	SetParameter(kParam_In2, kDefaultValue_In2 );
	SetParameter(kParam_LGain, kDefaultValue_LGain );
	SetParameter(kParam_RGain, kDefaultValue_RGain );
	SetParameter(kParam_MGain, kDefaultValue_MGain );
	SetParameter(kParam_SGain, kDefaultValue_SGain );
	SetParameter(kParam_Out1, kDefaultValue_Out1 );
	SetParameter(kParam_Out2, kDefaultValue_Out2 );
    
    lGainPrev = 1;
    rGainPrev = 1;
    mGainPrev = 1;
    sGainPrev = 1;
    
    #if AU_DEBUG_DISPATCHER
	mDebugDispatcher = new AUDebugDispatcher (this);
    #endif
    }


ComponentResult StereoImageMunger::GetParameterValueStrings(
        AudioUnitScope inScope,
        AudioUnitParameterID inParameterID,
        CFArrayRef *outStrings)
    {
    if (inScope != kAudioUnitScope_Global)
        return kAudioUnitErr_InvalidProperty;
    
    switch(inParameterID)
        {
        case kParam_In1:
        case kParam_In2:
        case kParam_Out1:
        case kParam_Out2:
            if(outStrings != NULL)
                {
                CFStringRef	strings[] =
                    {
                    kParameterValue_LName,
                    kParameterValue_RName,
                    kParameterValue_MName,
                    kParameterValue_SName
                    };
                *outStrings = CFArrayCreate( NULL, (const void **)strings, 4, NULL);
                }
            return noErr;
            
        default:
            return kAudioUnitErr_InvalidProperty;
        }
    }



//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//	StereoImageMunger::GetParameterInfo
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
ComponentResult	StereoImageMunger::GetParameterInfo(
        AudioUnitScope inScope,
        AudioUnitParameterID inParameterID,
        AudioUnitParameterInfo &outParameterInfo )
    {
	ComponentResult result = noErr;

	outParameterInfo.flags = kAudioUnitParameterFlag_IsWritable
						   | kAudioUnitParameterFlag_IsReadable;
    
    if(inScope == kAudioUnitScope_Global)
        {
        switch(inParameterID)
            {
            case kParam_In1:
                AUBase::FillInParameterName (outParameterInfo, kParameterNameIn1, false);
                outParameterInfo.unit = kAudioUnitParameterUnit_Indexed;
                outParameterInfo.minValue = 0;
                outParameterInfo.maxValue = 3;
                outParameterInfo.defaultValue = kDefaultValue_In1;
                outParameterInfo.flags |= kAudioUnitParameterFlag_NonRealTime;
                break;
            case kParam_In2:
                AUBase::FillInParameterName (outParameterInfo, kParameterNameIn2, false);
                outParameterInfo.unit = kAudioUnitParameterUnit_Indexed;
                outParameterInfo.minValue = 0;
                outParameterInfo.maxValue = 3;
                outParameterInfo.defaultValue = kDefaultValue_In2;
                outParameterInfo.flags |= kAudioUnitParameterFlag_NonRealTime;
                break;
            case kParam_LGain:
                AUBase::FillInParameterName (outParameterInfo, kParameterNameLGain, false);
                outParameterInfo.unit = kAudioUnitParameterUnit_Percent;
                outParameterInfo.minValue = -200;
                outParameterInfo.maxValue = 200;
                outParameterInfo.defaultValue = kDefaultValue_LGain;
                break;
            case kParam_RGain:
                AUBase::FillInParameterName (outParameterInfo, kParameterNameRGain, false);
                outParameterInfo.unit = kAudioUnitParameterUnit_Percent;
                outParameterInfo.minValue = -200;
                outParameterInfo.maxValue = 200;
                outParameterInfo.defaultValue = kDefaultValue_RGain;
                break;
            case kParam_MGain:
                AUBase::FillInParameterName (outParameterInfo, kParameterNameMGain, false);
                outParameterInfo.unit = kAudioUnitParameterUnit_Percent;
                outParameterInfo.minValue = -200;
                outParameterInfo.maxValue = 200;
                outParameterInfo.defaultValue = kDefaultValue_MGain;
                break;
            case kParam_SGain:
                AUBase::FillInParameterName (outParameterInfo, kParameterNameSGain, false);
                outParameterInfo.unit = kAudioUnitParameterUnit_Percent;
                outParameterInfo.minValue = -200;
                outParameterInfo.maxValue = 200;
                outParameterInfo.defaultValue = kDefaultValue_SGain;
                break;
            case kParam_Out1:
                AUBase::FillInParameterName (outParameterInfo, kParameterNameOut1, false);
                outParameterInfo.unit = kAudioUnitParameterUnit_Indexed;
                outParameterInfo.minValue = 0;
                outParameterInfo.maxValue = 3;
                outParameterInfo.defaultValue = kDefaultValue_Out1;
                outParameterInfo.flags |= kAudioUnitParameterFlag_NonRealTime;
                break;
            case kParam_Out2:
                AUBase::FillInParameterName (outParameterInfo, kParameterNameOut2, false);
                outParameterInfo.unit = kAudioUnitParameterUnit_Indexed;
                outParameterInfo.minValue = 0;
                outParameterInfo.maxValue = 3;
                outParameterInfo.defaultValue = kDefaultValue_Out2;
                outParameterInfo.flags |= kAudioUnitParameterFlag_NonRealTime;
                break;
            default:
                result = kAudioUnitErr_InvalidParameter;
                break;
            }
        }
	else
        result = kAudioUnitErr_InvalidParameter;

	return result;
    }

ComponentResult	StereoImageMunger::GetPropertyInfo(
        AudioUnitPropertyID	inID,
        AudioUnitScope inScope,
        AudioUnitElement inElement,
        UInt32 &outDataSize,
        Boolean &outWritable)
    {
	return AUEffectBase::GetPropertyInfo (inID, inScope, inElement, outDataSize, outWritable);
    }

ComponentResult	StereoImageMunger::GetProperty(
        AudioUnitPropertyID inID,
        AudioUnitScope inScope,
        AudioUnitElement inElement,
        void *outData )
    {
	return AUEffectBase::GetProperty (inID, inScope, inElement, outData);
    }

OSStatus StereoImageMunger::ProcessBufferLists(
        AudioUnitRenderActionFlags &ioActionFlags,
        const AudioBufferList &inBufferList,
        AudioBufferList &outBufferList,
        UInt32 inFramesToProcess )
    {
    // silent = ioActionFlags & kAudioUnitRenderAction_OutputIsSilence;
	
    const AudioBuffer *srcBuffer = inBufferList.mBuffers;
    AudioBuffer *destBuffer = outBufferList.mBuffers;
    
    const Float32 *src1 = (const Float32 *) srcBuffer[0].mData;
    const Float32 *src2 = (const Float32 *) srcBuffer[1 % inBufferList.mNumberBuffers].mData;
    Float32 *dst1 = (Float32 *) destBuffer[0].mData;
    Float32 *dst2 = (Float32 *) destBuffer[1 % outBufferList.mNumberBuffers].mData;
    
	int in1Mode   = (int) GetParameter( kParam_In1 );
	int in2Mode   = (int) GetParameter( kParam_In2 );
	Float32 lGain =       GetParameter( kParam_LGain ) / 100.0;
	Float32 rGain =       GetParameter( kParam_RGain ) / 100.0;
	Float32 mGain =       GetParameter( kParam_MGain ) / 100.0;
	Float32 sGain =       GetParameter( kParam_SGain ) / 100.0;
	int out1Mode  = (int) GetParameter( kParam_Out1 );
	int out2Mode  = (int) GetParameter( kParam_Out2 );
	
    Float32 inFramesToProcessFloat = inFramesToProcess;
    UInt32 frame;
	for (frame = 0; frame < inFramesToProcess; frame++, src1++, src2++, dst1++, dst2++)
        {
        Float32 w = (Float32) frame / inFramesToProcessFloat;
        Float32 w1 = 1.0 - w;

		Float32 i1 = *src1;
        Float32 i2 = *src2;

        // Normalize and apply LR processing
        
        Float32 l = custom2l(i1, i2, in1Mode, in2Mode) * (lGain * w + lGainPrev * w1);
        Float32 r = custom2r(i1, i2, in1Mode, in2Mode) * (rGain * w + rGainPrev * w1);
        
        // Apply MS processing
        
        Float32 m = (l + r) / 2.0 * (mGain * w + mGainPrev * w1);
        Float32 s = (l - r) / 2.0 * (sGain * w + sGainPrev * w1);
        
        // Convert to output modes
        
        *dst1 = ms2custom(m, s, out1Mode);
        *dst2 = ms2custom(m, s, out2Mode);
        }

    lGainPrev = lGain;
    rGainPrev = rGain;
    mGainPrev = mGain;
    sGainPrev = sGain;
	
	return noErr;
    }


static const Float32
    leftTable[4][4] =
    
        { { .5,  0,  0,  0 },
          {  1,  0,  2,  2 },
          {  1, -1, .5,  1 },
          {  1,  1,  1, .5 } },
          
    rightTable[4][4] =
    
        { {  0,  1,  2, -2 },
          {  0, .5,  0,  0 },
          { -1,  1, .5, -1 },
          {  1,  1,  1, -.5 } };

Float32 custom2l(Float32 i1, Float32 i2, int i1mode, int i2mode)
    {
    return leftTable[i1mode][i2mode] * i2
         + leftTable[i2mode][i1mode] * i1;
    }

Float32 custom2r(Float32 i1, Float32 i2, int i1mode, int i2mode)
    {
    return rightTable[i1mode][i2mode] * i2
         + rightTable[i2mode][i1mode] * i1;
    }

Float32 ms2custom(Float32 m, Float32 s, int mode)
    {
    switch(mode)
        {
        case kParameterValue_L:
            return m+s;
        case kParameterValue_R:
            return m-s;
        case kParameterValue_M:
            return m;
        case kParameterValue_S:
            return s;
        default:
            return 0;
        }
    }
