/*
*	File:		StereoImageMunger.h
*	
*	Version:	1.0.2
* 
*	Created:	2005/3/24
*	
*	Copyright:  Copyright © 2005 Paul Cantrell
*
*  http://innig.net/
*  
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
* 
*  (1) Redistributions of source code must retain the above copyright
*      notice, this list of conditions and the following disclaimer.
* 
*  (2) Redistributions in binary form must reproduce the above copyright
*      notice, this list of conditions and the following disclaimer in
*      the documentation and/or other materials provided with the
*      distribution. 
* 
*  (3) The name of the author may not be used to endorse or promote
*      products derived from this software without specific prior
*      written permission.
* 
*  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
*  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
*  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
*  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
*  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
*  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
*  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
*  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
*  OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include "AUEffectBase.h"
#include "StereoImageMungerVersion.h"

#if AU_DEBUG_DISPATCHER
	#include "AUDebugDispatcher.h"
#endif


#ifndef __StereoImageMunger_h__
#define __StereoImageMunger_h__


#pragma mark ____StereoImageMunger Parameters

// parameters
static const float kDefaultValue_In1   = 0;
static const float kDefaultValue_In2   = 1;
static const float kDefaultValue_LGain = 100;
static const float kDefaultValue_RGain = 100;
static const float kDefaultValue_MGain = 100;
static const float kDefaultValue_SGain = 100;
static const float kDefaultValue_Out1  = 0;
static const float kDefaultValue_Out2  = 1;

static CFStringRef kParameterNameIn1   = CFSTR("In 1");
static CFStringRef kParameterNameIn2   = CFSTR("In 2");
static CFStringRef kParameterNameLGain = CFSTR("L gain");
static CFStringRef kParameterNameRGain = CFSTR("R gain");
static CFStringRef kParameterNameMGain = CFSTR("M gain");
static CFStringRef kParameterNameSGain = CFSTR("S gain");
static CFStringRef kParameterNameOut1  = CFSTR("Out 1");
static CFStringRef kParameterNameOut2  = CFSTR("Out 2");

static const int kParameterValue_L = 0;
static const int kParameterValue_R = 1;
static const int kParameterValue_M = 2;
static const int kParameterValue_S = 3;

static CFStringRef kParameterValue_LName = CFSTR( "Left" );
static CFStringRef kParameterValue_RName = CFSTR( "Right" );
static CFStringRef kParameterValue_MName = CFSTR( "Mid (Sum)" );
static CFStringRef kParameterValue_SName = CFSTR( "Side (Difference)" );

enum
    {
	kParam_In1 = 0,
	kParam_In2,
	kParam_LGain,
	kParam_RGain,
	kParam_MGain,
	kParam_SGain,
	kParam_Out1,
	kParam_Out2,

	kNumberOfParameters
    };

#pragma mark ____StereoImageMunger

class StereoImageMunger : public AUEffectBase
    {
    public:
        StereoImageMunger(AudioUnit component);
        #if AU_DEBUG_DISPATCHER
        virtual ~StereoImageMunger () { delete mDebugDispatcher; }
        #endif
        
        virtual	ComponentResult		GetParameterValueStrings(AudioUnitScope			inScope,
                                                             AudioUnitParameterID	inParameterID,
                                                             CFArrayRef *			outStrings);
        
        virtual	ComponentResult		GetParameterInfo(AudioUnitScope			inScope,
                                                     AudioUnitParameterID	inParameterID,
                                                     AudioUnitParameterInfo	&outParameterInfo);
        
        virtual ComponentResult		GetPropertyInfo(AudioUnitPropertyID		inID,
                                                    AudioUnitScope			inScope,
                                                    AudioUnitElement		inElement,
                                                    UInt32 &			outDataSize,
                                                    Boolean	&			outWritable );
        
        virtual ComponentResult		GetProperty(AudioUnitPropertyID inID,
                                                AudioUnitScope 		inScope,
                                                AudioUnitElement 		inElement,
                                                void *			outData);
        
        virtual	bool				SupportsTail () { return false; }
        
        virtual ComponentResult	Version() { return kStereoImageMungerVersion; }
        
        virtual OSStatus            ProcessBufferLists(
                                                AudioUnitRenderActionFlags &	ioActionFlags,
                                                const AudioBufferList &			inBuffer,
                                                AudioBufferList &				outBuffer,
                                                UInt32							inFramesToProcess );
        
    private:
        Float32 lGainPrev, rGainPrev, mGainPrev, sGainPrev;
    };


// Utilities

Float32 custom2l(Float32 i1, Float32 i2, int i1mode, int i2mode);
Float32 custom2r(Float32 i1, Float32 i2, int i1mode, int i2mode);

Float32 ms2custom(Float32 m, Float32 s, int mode);


#endif
